@@ -289,33 +289,22 @@ static void print_usage(void)
 	Erase();
 }
 
-/* ------------------------------------------------------------------------ */
-/* get_block_type                                                           */
-/* returns: metatype if block is a GFS2 structure block type                */
-/*          0 if block is not a GFS2 structure                              */
-/* ------------------------------------------------------------------------ */
-uint32_t get_block_type(const struct gfs2_buffer_head *lbh, int *structlen)
+const struct lgfs2_metadata *get_block_type(char *buf)
 {
-	uint32_t ty = lgfs2_get_block_type(lbh);
+	uint32_t t = lgfs2_get_block_type(buf);
 
-	if (ty != 0 && structlen != NULL) {
-		unsigned ver = sbd.gfs1 ? LGFS2_MD_GFS1 : LGFS2_MD_GFS2;
-		const struct lgfs2_metadata *mtype = lgfs2_find_mtype(ty, ver);
-		if (mtype != NULL)
-			*structlen = mtype->size;
-		else
-			*structlen = sbd.bsize;
-	}
-	return ty;
+	if (t != 0)
+		return lgfs2_find_mtype(t, sbd.gfs1 ? LGFS2_MD_GFS1 : LGFS2_MD_GFS2);
+	return NULL;
 }
 
-/* ------------------------------------------------------------------------ */
-/* display_block_type                                                       */
-/* returns: metatype if block is a GFS2 structure block type                */
-/*          0 if block is not a GFS2 structure                              */
-/* ------------------------------------------------------------------------ */
-int display_block_type(struct gfs2_buffer_head *dbh, int from_restore)
+/**
+ * returns: metatype if block is a GFS2 structure block type
+ *          0 if block is not a GFS2 structure
+ */
+int display_block_type(char *buf, uint64_t addr, int from_restore)
 {
+	const struct lgfs2_metadata *mtype;
 	const struct gfs2_meta_header *mh;
 	int ret_type = 0; /* return type */
 
@@ -334,7 +323,7 @@ int display_block_type(struct gfs2_buffer_head *dbh, int from_restore)
 	else if (block == JOURNALS_DUMMY_BLOCK)
 		print_gfs2("Journal Status:      ");
 	else
-		print_gfs2("%"PRIu64"    (0x%"PRIx64")", dbh->b_blocknr, dbh->b_blocknr);
+		print_gfs2("%"PRIu64"    (0x%"PRIx64")", addr, addr);
 	if (termlines) {
 		if (edit_row[dmode] == -1)
 			COLORS_NORMAL;
@@ -350,56 +339,17 @@ int display_block_type(struct gfs2_buffer_head *dbh, int from_restore)
 		ret_type = GFS2_METATYPE_DI;
 		struct_len = 0;
 	} else {
-		ret_type = get_block_type(dbh, &struct_len);
-		switch (ret_type) {
-		case GFS2_METATYPE_SB:   /* 1 */
-			print_gfs2("(superblock)");
-			break;
-		case GFS2_METATYPE_RG:   /* 2 */
-			print_gfs2("(rsrc grp hdr)");
-			break;
-		case GFS2_METATYPE_RB:   /* 3 */
-			print_gfs2("(rsrc grp bitblk)");
-			break;
-		case GFS2_METATYPE_DI:   /* 4 */
-			print_gfs2("(disk inode)");
-			break;
-		case GFS2_METATYPE_IN:   /* 5 */
-			print_gfs2("(indir blklist)");
-			break;
-		case GFS2_METATYPE_LF:   /* 6 */
-			print_gfs2("(directory leaf)");
-			break;
-		case GFS2_METATYPE_JD:
-			print_gfs2("(journal data)");
-			break;
-		case GFS2_METATYPE_LH:
-			print_gfs2("(log header)");
-			break;
-		case GFS2_METATYPE_LD:
-		 	print_gfs2("(log descriptor)");
-			break;
-		case GFS2_METATYPE_EA:
-			print_gfs2("(extended attr hdr)");
-			break;
-		case GFS2_METATYPE_ED:
-			print_gfs2("(extended attr data)");
-			break;
-		case GFS2_METATYPE_LB:
-			print_gfs2("(log buffer)");
-			break;
-		case GFS2_METATYPE_QC:
-			print_gfs2("(quota change)");
-			break;
-		case 0:
+		mtype = get_block_type(buf);
+		if (mtype != NULL) {
+			print_gfs2("(%s)", mtype->display);
+			struct_len = mtype->size;
+			ret_type = mtype->mh_type;
+		} else {
 			struct_len = sbd.bsize;
-			break;
-		default:
-			print_gfs2("(wtf?)");
-			break;
+			ret_type = 0;
 		}
 	}
-	mh = dbh->iov.iov_base;
+	mh = (void *)buf;
 	eol(0);
 	if (from_restore)
 		return ret_type;
@@ -485,7 +435,7 @@ int display_block_type(struct gfs2_buffer_head *dbh, int from_restore)
 			}
 		}
 		if (rgd)
-			gfs2_rgrp_relse(rgd);
+			gfs2_rgrp_relse(&sbd, rgd);
  	}
 	if (block == sbd.sd_sb.sb_root_dir.no_addr)
 		print_gfs2("--------------- Root directory ------------------");
@@ -531,20 +481,6 @@ int display_block_type(struct gfs2_buffer_head *dbh, int from_restore)
 	return ret_type;
 }
 
-static const struct lgfs2_metadata *find_mtype(uint32_t mtype, const unsigned versions)
-{
-	const struct lgfs2_metadata *m = lgfs2_metadata;
-	unsigned n = 0;
-
-	do {
-		if ((m[n].versions & versions) && m[n].mh_type == mtype)
-			return &m[n];
-		n++;
-	} while (n < lgfs2_metadata_size);
-
-	return NULL;
-}
-
 static int get_pnum(int ptroffset)
 {
 	int pnum;
@@ -568,7 +504,7 @@ static int hexdump(uint64_t startaddr, int len, int trunc_zeros,
 	const char *lpBuffer = bh->b_data;
 	const char *zeros_strt = lpBuffer + sbd.bsize;
 	int print_field, cursor_line;
-	const uint32_t block_type = get_block_type(bh, NULL);
+	const struct lgfs2_metadata *m = get_block_type(bh->b_data);
 	uint64_t *ref;
 	int ptroffset = 0;
 
@@ -673,8 +609,6 @@ static int hexdump(uint64_t startaddr, int len, int trunc_zeros,
 		}
 		print_gfs2("] ");
 		if (print_field >= 0) {
-			const struct lgfs2_metadata *m = find_mtype(block_type,
-			               sbd.gfs1 ? LGFS2_MD_GFS1 : LGFS2_MD_GFS2);
 			if (m) {
 				const struct lgfs2_metafield *f;
 				unsigned n;
@@ -689,7 +623,9 @@ static int hexdump(uint64_t startaddr, int len, int trunc_zeros,
 			}
 
 		}
-		if (cursor_line) {
+		if (m && cursor_line) {
+			const uint32_t block_type = m->mh_type;
+
 			if (block_type == GFS2_METATYPE_IN ||
 			    block_type == GFS2_METATYPE_LD ||
 			    ((block_type == GFS2_METATYPE_DI) &&
@@ -723,7 +659,7 @@ static int hexdump(uint64_t startaddr, int len, int trunc_zeros,
 		if ((const char *)pointer >= zeros_strt)
 			break;
 	} /* while */
-	if (block_type == GFS2_METATYPE_LD && ptroffset >= struct_len) {
+	if (m && m->mh_type == GFS2_METATYPE_LD && ptroffset >= struct_len) {
 		COLORS_NORMAL;
 		eol(0);
 		print_gfs2("         * 'j' will jump to the journaled block, "
@@ -761,7 +697,7 @@ static void rgcount(void)
 	       (unsigned long long)sbd.md.riinode->i_di.di_size /
 	       sizeof(struct gfs2_rindex));
 	inode_put(&sbd.md.riinode);
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
 	exit(EXIT_SUCCESS);
 }
 
@@ -850,7 +786,7 @@ static void set_rgrp_flags(int rgnum, uint32_t new_flags, int modify, int full)
 	rgblk = get_rg_addr(rgnum);
 	rbh = bread(&sbd, rgblk);
 	if (sbd.gfs1)
-		gfs_rgrp_in(&rg.rg1, rbh);
+		gfs_rgrp_in(&rg.rg1, rbh->b_data);
 	else
 		gfs2_rgrp_in(&rg.rg2, rbh->b_data);
 	if (modify) {
@@ -859,7 +795,7 @@ static void set_rgrp_flags(int rgnum, uint32_t new_flags, int modify, int full)
 		       (unsigned long long)rgblk, rg.rg2.rg_flags, new_flags);
 		rg.rg2.rg_flags = new_flags;
 		if (sbd.gfs1)
-			gfs_rgrp_out(&rg.rg1, rbh);
+			gfs_rgrp_out(&rg.rg1, rbh->b_data);
 		else
 			gfs2_rgrp_out(&rg.rg2, rbh->b_data);
 		bmodified(rbh);
@@ -1057,7 +993,7 @@ static int read_master_dir(void)
 	if (bh == NULL)
 		return 1;
 	gfs2_dinode_in(&di, bh->b_data);
-	do_dinode_extended(&di, bh); /* get extended data, if any */
+	do_dinode_extended(&di, bh->b_data); /* get extended data, if any */
 	memcpy(&masterdir, &indirect[0], sizeof(struct indirect_info));
 	return 0;
 }
@@ -1098,7 +1034,7 @@ int display(int identify_only, int trunc_zeros, uint64_t flagref,
 		}
 	}
 	line = 1;
-	gfs2_struct_type = display_block_type(bh, FALSE);
+	gfs2_struct_type = display_block_type(bh->b_data, bh->b_blocknr, FALSE);
 	if (identify_only)
 		return 0;
 	indirect_blocks = 0;
@@ -1128,7 +1064,7 @@ int display(int identify_only, int trunc_zeros, uint64_t flagref,
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_DI) {
 		gfs2_dinode_in(&di, bh->b_data);
-		do_dinode_extended(&di, bh); /* get extended data, if any */
+		do_dinode_extended(&di, bh->b_data); /* get extended data, if any */
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_IN) { /* indirect block list */
 		if (blockhist) {
@@ -1163,7 +1099,7 @@ int display(int identify_only, int trunc_zeros, uint64_t flagref,
 		        flagref, ref_blk);
 	else if (dmode == GFS2_MODE) { /* if structure display */
 		if (block != JOURNALS_DUMMY_BLOCK)
-			display_gfs2(bh);  /* display the gfs2 structure */
+			display_gfs2(bh->b_data);
 	} else
 		display_extended();        /* display extended blocks       */
 	/* No else here because display_extended can switch back to hex mode */
@@ -1256,7 +1192,7 @@ static uint64_t find_metablockoftype_slow(uint64_t startblk, int metatype, int p
 		else
 			printf("%llu\n", (unsigned long long)blk);
 	}
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
 	if (print)
 		exit(0);
 	return blk;
@@ -1275,7 +1211,7 @@ static int find_rg_metatype(struct rgrp_tree *rgd, uint64_t *blk, uint64_t start
 		for (j = 0; j < m; j++) {
 			*blk = ibuf[j];
 			bhp = bread(&sbd, *blk);
-			found = (*blk > startblk) && !gfs2_check_meta(bhp, mtype);
+			found = (*blk > startblk) && !gfs2_check_meta(bhp->b_data, mtype);
 			brelse(bhp);
 			if (found) {
 				free(ibuf);
@@ -1320,7 +1256,7 @@ static uint64_t find_metablockoftype_rg(uint64_t startblk, int metatype, int pri
 	if (!rgd) {
 		if (print)
 			printf("0\n");
-		gfs2_rgrp_free(&sbd.rgtree);
+		gfs2_rgrp_free(&sbd, &sbd.rgtree);
 		if (print)
 			exit(-1);
 	}
@@ -1334,7 +1270,7 @@ static uint64_t find_metablockoftype_rg(uint64_t startblk, int metatype, int pri
 		if (found)
 			break;
 
-		gfs2_rgrp_relse(rgd);
+		gfs2_rgrp_relse(&sbd, rgd);
 	}
 
 	if (!found)
@@ -1345,7 +1281,7 @@ static uint64_t find_metablockoftype_rg(uint64_t startblk, int metatype, int pri
 		else
 			printf("%llu\n", (unsigned long long)blk);
 	}
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
 	if (print)
 		exit(0);
 	return blk;
@@ -1379,7 +1315,7 @@ static uint64_t find_metablockoftype(const char *strtype, int print)
 			"specified: must be one of:\n");
 		fprintf(stderr, "sb rg rb di in lf jd lh ld"
 			" ea ed lb 13 qc\n");
-		gfs2_rgrp_free(&sbd.rgtree);
+		gfs2_rgrp_free(&sbd, &sbd.rgtree);
 		exit(-1);
 	}
 	return blk;
@@ -1637,8 +1573,12 @@ static void jump(void)
 	if (dmode == HEX_MODE) {
 		unsigned int col2;
 		uint64_t *b;
-		const uint32_t block_type = get_block_type(bh, NULL);
-		
+		const struct lgfs2_metadata *mtype = get_block_type(bh->b_data);
+		uint32_t block_type = 0;
+
+		if (mtype != NULL)
+			block_type = mtype->mh_type;
+
 		/* special exception for log descriptors: jump the journaled
 		   version of the block, not the "real" block */
 		if (block_type == GFS2_METATYPE_LD) {
@@ -1667,35 +1607,28 @@ static void jump(void)
 	}
 }
 
-/* ------------------------------------------------------------------------ */
-/* print block type                                                         */
-/* ------------------------------------------------------------------------ */
-static void print_block_type(uint64_t tblock, int type, const char *additional)
+static void print_block_type(uint64_t tblock, const struct lgfs2_metadata *type)
 {
-	if (type <= GFS2_METATYPE_QC)
-		printf("%d (Block %lld is type %d: %s%s)\n", type,
-		       (unsigned long long)tblock, type, block_type_str[type],
-		       additional);
+	if (type != NULL && type->nfields > 0)
+		printf("%d (Block %"PRIu64" is type %d: %s)\n", type->mh_type,
+		       tblock, type->mh_type, type->display);
 	else
-		printf("%d (Block %lld is type %d: unknown%s)\n", type,
-		       (unsigned long long)tblock, type, additional);
+		printf("%d (Block %"PRIu64" is type %d: unknown)\n", type->mh_type,
+		       tblock, type->mh_type);
 }
 
-/* ------------------------------------------------------------------------ */
-/* find_print block type                                                    */
-/* ------------------------------------------------------------------------ */
 static void find_print_block_type(void)
 {
 	uint64_t tblock;
 	struct gfs2_buffer_head *lbh;
-	int type;
+	const struct lgfs2_metadata *type;
 
 	tblock = blockstack[blockhist % BLOCK_STACK_SIZE].block;
 	lbh = bread(&sbd, tblock);
-	type = get_block_type(lbh, NULL);
-	print_block_type(tblock, type, "");
+	type = get_block_type(lbh->b_data);
+	print_block_type(tblock, type);
 	brelse(lbh);
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
 	exit(0);
 }
 
@@ -1738,7 +1671,7 @@ static void find_print_block_rg(int bitmap)
 			printf("-1 (block invalid or part of an rgrp).\n");
 		}
 	}
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
 	exit(0);
 }
 
@@ -1759,7 +1692,7 @@ static void find_change_block_alloc(int *newval)
 		       *newval);
 		for (i = GFS2_BLKST_FREE; i <= GFS2_BLKST_DINODE; i++)
 			printf("%d - %s\n", i, allocdesc[sbd.gfs1][i]);
-		gfs2_rgrp_free(&sbd.rgtree);
+		gfs2_rgrp_free(&sbd, &sbd.rgtree);
 		exit(-1);
 	}
 	ablock = blockstack[blockhist % BLOCK_STACK_SIZE].block;
@@ -1783,14 +1716,14 @@ static void find_change_block_alloc(int *newval)
 				}
 				printf("%d (%s)\n", type, allocdesc[sbd.gfs1][type]);
 			}
-			gfs2_rgrp_relse(rgd);
+			gfs2_rgrp_relse(&sbd, rgd);
 		} else {
-			gfs2_rgrp_free(&sbd.rgtree);
+			gfs2_rgrp_free(&sbd, &sbd.rgtree);
 			printf("-1 (block invalid or part of an rgrp).\n");
 			exit(-1);
 		}
 	}
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
 	if (newval)
 		fsync(sbd.device_fd);
 	exit(0);
@@ -1803,17 +1736,15 @@ static void process_field(const char *field, const char *nstr)
 {
 	uint64_t fblock;
 	struct gfs2_buffer_head *rbh;
-	int type;
 	const struct lgfs2_metadata *mtype;
 	const struct lgfs2_metafield *mfield;
 
 	fblock = blockstack[blockhist % BLOCK_STACK_SIZE].block;
 	rbh = bread(&sbd, fblock);
-	type = get_block_type(rbh, NULL);
-
-	mtype = lgfs2_find_mtype(type, sbd.gfs1 ? LGFS2_MD_GFS1 : LGFS2_MD_GFS2);
+	mtype = get_block_type(rbh->b_data);
 	if (mtype == NULL) {
-		fprintf(stderr, "Metadata type '%d' invalid\n", type);
+		fprintf(stderr, "Metadata type of block %"PRIx64" not recognised\n",
+		        fblock);
 		exit(1);
 	}
 
@@ -2116,12 +2047,11 @@ static void interactive_mode(void)
 /* ------------------------------------------------------------------------ */
 /* gfs_log_header_in - read in a gfs1-style log header                      */
 /* ------------------------------------------------------------------------ */
-void gfs_log_header_in(struct gfs_log_header *head,
-		       struct gfs2_buffer_head *lbh)
+void gfs_log_header_in(struct gfs_log_header *head, const char *buf)
 {
-	struct gfs_log_header *str = lbh->iov.iov_base;
+	const struct gfs_log_header *str = (void *)buf;
 
-	gfs2_meta_header_in(&head->lh_header, lbh->b_data);
+	gfs2_meta_header_in(&head->lh_header, buf);
 
 	head->lh_flags = be32_to_cpu(str->lh_flags);
 	head->lh_pad = be32_to_cpu(str->lh_pad);
@@ -2280,7 +2210,7 @@ static int count_dinode_blks(struct rgrp_tree *rgd, int bitmap,
 			    rgd->bits[bitmap].bi_start + b);
 		byte = rbh->b_data + off + (b / GFS2_NBBY);
 		bit = (b % GFS2_NBBY) * GFS2_BIT_SIZE;
-		if (gfs2_check_meta(tbh, GFS2_METATYPE_DI) == 0) {
+		if (gfs2_check_meta(tbh->b_data, GFS2_METATYPE_DI) == 0) {
 			dinodes++;
 			new_state = GFS2_BLKST_DINODE;
 		} else {
@@ -2330,7 +2260,7 @@ static void rg_repair(void)
 	for (n = osi_first(&sbd.rgtree); n; n = osi_next(n)) {
 		rgd = (struct rgrp_tree *)n;
 		if (gfs2_rgrp_read(&sbd, rgd) == 0) { /* was read in okay */
-			gfs2_rgrp_relse(rgd);
+			gfs2_rgrp_relse(&sbd, rgd);
 			continue; /* ignore it */
 		}
 		/* If we get here, it's because we have an rgrp in the rindex
@@ -2351,7 +2281,7 @@ static void rg_repair(void)
 
 			printf("Bitmap #%d:", b);
 			rbh = bread(&sbd, rgd->ri.ri_addr + b);
-			if (gfs2_check_meta(rbh, mtype)) { /* wrong type */
+			if (gfs2_check_meta(rbh->b_data, mtype)) { /* wrong type */
 				printf("Damaged. Repairing...");
 				/* Fix the meta header */
 				memset(rbh->b_data, 0, sbd.bsize);
@@ -2522,7 +2452,7 @@ static void process_parameters(int argc, char *argv[], int pass)
 				printf("Error: field not specified.\n");
 				printf("Format is: %s -p <block> field "
 				       "<field> [newvalue]\n", argv[0]);
-				gfs2_rgrp_free(&sbd.rgtree);
+				gfs2_rgrp_free(&sbd, &sbd.rgtree);
 				exit(EXIT_FAILURE);
 			}
 			process_field(argv[i], argv[i + 1]);
@@ -2555,7 +2485,7 @@ static void process_parameters(int argc, char *argv[], int pass)
 				printf("Error: rg # not specified.\n");
 				printf("Format is: %s rgflags rgnum"
 				       "[newvalue]\n", argv[0]);
-				gfs2_rgrp_free(&sbd.rgtree);
+				gfs2_rgrp_free(&sbd, &sbd.rgtree);
 				exit(EXIT_FAILURE);
 			}
 			if (argv[i][0]=='0' && argv[i][1]=='x')
@@ -2572,16 +2502,16 @@ static void process_parameters(int argc, char *argv[], int pass)
 					new_flags = atoi(argv[i]);
 			}
 			set_rgrp_flags(rg, new_flags, set, FALSE);
-			gfs2_rgrp_free(&sbd.rgtree);
+			gfs2_rgrp_free(&sbd, &sbd.rgtree);
 			exit(EXIT_SUCCESS);
 		} else if (!strcmp(argv[i], "rg")) {
 			int rg;
-				
+
 			i++;
 			if (i >= argc - 1) {
 				printf("Error: rg # not specified.\n");
 				printf("Format is: %s rg rgnum\n", argv[0]);
-				gfs2_rgrp_free(&sbd.rgtree);
+				gfs2_rgrp_free(&sbd, &sbd.rgtree);
 				exit(EXIT_FAILURE);
 			}
 			rg = atoi(argv[i]);
@@ -2590,7 +2520,7 @@ static void process_parameters(int argc, char *argv[], int pass)
 				push_block(temp_blk);
 			} else {
 				set_rgrp_flags(rg, 0, FALSE, TRUE);
-				gfs2_rgrp_free(&sbd.rgtree);
+				gfs2_rgrp_free(&sbd, &sbd.rgtree);
 				exit(EXIT_SUCCESS);
 			}
 		} else if (!strcmp(argv[i], "rgbitmaps")) {
@@ -2603,7 +2533,7 @@ static void process_parameters(int argc, char *argv[], int pass)
 				printf("Error: rg # not specified.\n");
 				printf("Format is: %s rgbitmaps rgnum\n",
 				       argv[0]);
-				gfs2_rgrp_free(&sbd.rgtree);
+				gfs2_rgrp_free(&sbd, &sbd.rgtree);
 				exit(EXIT_FAILURE);
 			}
 			rg = atoi(argv[i]);
@@ -2611,7 +2541,7 @@ static void process_parameters(int argc, char *argv[], int pass)
 			rgd = gfs2_blk2rgrpd(&sbd, rgblk);
 			if (rgd == NULL) {
 				printf("Error: rg # is invalid.\n");
-				gfs2_rgrp_free(&sbd.rgtree);
+				gfs2_rgrp_free(&sbd, &sbd.rgtree);
 				exit(EXIT_FAILURE);
 			}
 			for (bmap = 0; bmap < rgd->ri.ri_length; bmap++)
@@ -2717,6 +2647,6 @@ int main(int argc, char *argv[])
 	close(fd);
 	if (indirect)
 		free(indirect);
-	gfs2_rgrp_free(&sbd.rgtree);
+	gfs2_rgrp_free(&sbd, &sbd.rgtree);
  	exit(EXIT_SUCCESS);
 }
