@@ -104,17 +104,16 @@ static void refresh_rgrp(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
 	log_debug(_("Block is part of rgrp 0x%llx; refreshing the rgrp.\n"),
 		  (unsigned long long)rgd->ri.ri_addr);
 	for (i = 0; i < rgd->ri.ri_length; i++) {
-		if (rgd->bits[i].bi_bh->b_blocknr != blkno)
+		if (rgd->ri.ri_addr + i != blkno)
 			continue;
 
-		memcpy(rgd->bits[i].bi_bh->b_data, bh->b_data, sdp->bsize);
-		bmodified(rgd->bits[i].bi_bh);
+		memcpy(rgd->bits[i].bi_data, bh->b_data, sdp->bsize);
+		rgd->bits[i].bi_modified = 1;
 		if (i == 0) { /* this is the rgrp itself */
 			if (sdp->gfs1)
-				gfs_rgrp_in((struct gfs_rgrp *)&rgd->rg,
-					    rgd->bits[0].bi_bh);
+				gfs_rgrp_in((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_data);
 			else
-				gfs2_rgrp_in(&rgd->rg, rgd->bits[0].bi_bh->b_data);
+				gfs2_rgrp_in(&rgd->rg, rgd->bits[0].bi_data);
 		}
 		break;
 	}
@@ -211,7 +210,7 @@ static int revoke_lo_scan_elements(struct gfs2_inode *ip, unsigned int start,
 			return error;
 
 		if (!first) {
-			if (gfs2_check_meta(bh, GFS2_METATYPE_LB))
+			if (gfs2_check_meta(bh->b_data, GFS2_METATYPE_LB))
 				continue;
 		}
 		while (offset + sizeof(uint64_t) <= sdp->sd_sb.sb_bsize) {
@@ -355,7 +354,7 @@ static int foreach_descriptor(struct gfs2_inode *ip, unsigned int start,
 			bmodified(bh);
 			brelse(bh);
 			return error;
-		} else if (gfs2_check_meta(bh, GFS2_METATYPE_LD)) {
+		} else if (gfs2_check_meta(bh->b_data, GFS2_METATYPE_LD)) {
 			bmodified(bh);
 			brelse(bh);
 			return -EIO;
@@ -636,11 +635,11 @@ static int rangecheck_jblock(struct gfs2_inode *ip, uint64_t block)
 	return meta_is_good;
 }
 
-static int rangecheck_jmeta(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, int h,
-			       int *is_valid, int *was_duplicate,
-			       void *private)
+static int rangecheck_jmeta(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                            int *is_valid, int *was_duplicate, void *private)
 {
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
 	int rc;
 
 	*bh = NULL;
@@ -649,7 +648,7 @@ static int rangecheck_jmeta(struct gfs2_inode *ip, uint64_t block,
 	rc = rangecheck_jblock(ip, block);
 	if (rc == meta_is_good) {
 		*bh = bread(ip->i_sbd, block);
-		*is_valid = (gfs2_check_meta(*bh, GFS2_METATYPE_IN) == 0);
+		*is_valid = (gfs2_check_meta((*bh)->b_data, GFS2_METATYPE_IN) == 0);
 		if (!(*is_valid)) {
 			log_err( _("Journal at block %lld (0x%llx) has a bad "
 				   "indirect block pointer %lld (0x%llx) "
