@@ -40,9 +40,14 @@ int check_n_fix_bitmap(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
 		{"free", "data", "unlinked", "inode", "reserved"},
 		/* gfs1 descriptions: */
 		{"free", "data", "free meta", "metadata", "reserved"}};
+	static struct rgrp_tree *prevrgd = NULL;
 
-	if (rgd == NULL || !rgrp_contains_block(rgd, blk))
+	if (prevrgd && rgrp_contains_block(prevrgd, blk)) {
+		rgd = prevrgd;
+	} else if (rgd == NULL || !rgrp_contains_block(rgd, blk)) {
 		rgd = gfs2_blk2rgrpd(sdp, blk);
+		prevrgd = rgd;
+	}
 
 	gfs1rg = (struct gfs_rgrp *)&rgd->rg;
 
@@ -150,9 +155,10 @@ int check_n_fix_bitmap(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
 	}
 	if (rewrite_rgrp) {
 		if (sdp->gfs1)
-			gfs_rgrp_out((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_bh);
+			gfs_rgrp_out((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_data);
 		else
-			gfs2_rgrp_out_bh(&rgd->rg, rgd->bits[0].bi_bh);
+			gfs2_rgrp_out(&rgd->rg, rgd->bits[0].bi_data);
+		rgd->bits[0].bi_modified = 1;
 	}
 	log_err( _("The bitmap was fixed.\n"));
 	return 0;
@@ -286,7 +292,7 @@ struct gfs2_inode *fsck_inode_get(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
 		return sysip;
 
 	if (sdp->gfs1)
-		ip = lgfs2_gfs_inode_get(sdp, bh);
+		ip = lgfs2_gfs_inode_get(sdp, bh->b_data);
 	else
 		ip = lgfs2_inode_get(sdp, bh);
 	if (ip)
@@ -554,7 +560,7 @@ int check_leaf(struct gfs2_inode *ip, int lindex, struct metawalk_fxns *pass,
 	/* Try to read in the leaf block. */
 	lbh = bread(sdp, *leaf_no);
 	/* Make sure it's really a valid leaf block. */
-	if (gfs2_check_meta(lbh, GFS2_METATYPE_LF)) {
+	if (gfs2_check_meta(lbh->b_data, GFS2_METATYPE_LF)) {
 		msg = _("that is not really a leaf");
 		goto bad_leaf;
 	}
@@ -577,12 +583,13 @@ int check_leaf(struct gfs2_inode *ip, int lindex, struct metawalk_fxns *pass,
 	   to use cpu_to_be32(), but we should check for incorrect values and
 	   replace them with the correct value. */
 
-	gfs2_leaf_in(leaf, lbh);
+	gfs2_leaf_in(leaf, lbh->b_data);
 	if (leaf->lf_dirent_format == (GFS2_FORMAT_DE << 16)) {
 		log_debug( _("incorrect lf_dirent_format at leaf #%" PRIu64
 			     "\n"), *leaf_no);
 		leaf->lf_dirent_format = GFS2_FORMAT_DE;
-		gfs2_leaf_out(leaf, lbh);
+		gfs2_leaf_out(leaf, lbh->b_data);
+		bmodified(lbh);
 		log_debug( _("Fixing lf_dirent_format.\n"));
 	}
 
@@ -617,7 +624,7 @@ int check_leaf(struct gfs2_inode *ip, int lindex, struct metawalk_fxns *pass,
 		   changed it. */
 		brelse(lbh);
 		lbh = bread(sdp, *leaf_no);
-		gfs2_leaf_in(leaf, lbh);
+		gfs2_leaf_in(leaf, lbh->b_data);
 		if (count != leaf->lf_entries) {
 			log_err( _("Leaf %llu (0x%llx) entry count in "
 				   "directory %llu (0x%llx) does not match "
@@ -629,11 +636,12 @@ int check_leaf(struct gfs2_inode *ip, int lindex, struct metawalk_fxns *pass,
 				 leaf->lf_entries, count);
 			if (query( _("Update leaf entry count? (y/n) "))) {
 				leaf->lf_entries = count;
-				gfs2_leaf_out(leaf, lbh);
+				gfs2_leaf_out(leaf, lbh->b_data);
+				bmodified(lbh);
 				log_warn( _("Leaf entry count updated\n"));
 			} else
 				log_err( _("Leaf entry count left in "
-					   "inconsistant state\n"));
+					   "inconsistent state\n"));
 		}
 	}
 out:
@@ -763,7 +771,7 @@ int check_leaf_blks(struct gfs2_inode *ip, struct metawalk_fxns *pass)
 		if (valid_block_ip(ip, leaf_no)) {
 			lbh = bread(sdp, leaf_no);
 			/* Make sure it's really a valid leaf block. */
-			if (gfs2_check_meta(lbh, GFS2_METATYPE_LF) == 0) {
+			if (gfs2_check_meta(lbh->b_data, GFS2_METATYPE_LF) == 0) {
 				brelse(lbh);
 				first_ok_leaf = leaf_no;
 				break;
@@ -1202,6 +1210,53 @@ static void file_ra(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 			      extlen * sdp->bsize, POSIX_FADV_WILLNEED);
 }
 
+static int do_check_metalist(struct iptr iptr, int height, struct gfs2_buffer_head **bhp,
+                             struct metawalk_fxns *pass)
+{
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
+	int was_duplicate = 0;
+	int is_valid = 1;
+	int error;
+
+	if (pass->check_metalist == NULL)
+		return 0;
+
+	error = pass->check_metalist(iptr, bhp, height, &is_valid,
+				     &was_duplicate, pass->private);
+	if (error == meta_error) {
+		stack;
+		log_info("\n");
+		log_info(_("Serious metadata error on block %"PRIu64" (0x%"PRIx64").\n"),
+		         block, block);
+		return error;
+	}
+	if (error == meta_skip_further) {
+		log_info("\n");
+		log_info(_("Unrecoverable metadata error on block %"PRIu64" (0x%"PRIx64")\n"),
+		         block, block);
+		log_info(_("Further metadata will be skipped.\n"));
+		return error;
+	}
+	if (!is_valid) {
+		log_debug("Skipping rejected block %"PRIu64" (0x%"PRIx64")\n", block, block);
+		if (pass->invalid_meta_is_fatal)
+			return meta_error;
+		return meta_skip_one;
+	}
+	if (was_duplicate) {
+		log_debug("Skipping duplicate %"PRIu64" (0x%"PRIx64")\n", block, block);
+		return meta_skip_one;
+	}
+	if (!valid_block_ip(ip, block)) {
+		log_debug("Skipping invalid block %"PRIu64" (0x%"PRIx64")\n", block, block);
+		if (pass->invalid_meta_is_fatal)
+			return meta_error;
+		return meta_skip_one;
+	}
+	return error;
+}
+
 /**
  * build_and_check_metalist - check a bunch of indirect blocks
  *                            This includes hash table blocks for directories
@@ -1217,12 +1272,13 @@ static int build_and_check_metalist(struct gfs2_inode *ip, osi_list_t *mlp,
 				    struct metawalk_fxns *pass)
 {
 	uint32_t height = ip->i_di.di_height;
-	struct gfs2_buffer_head *bh, *nbh, *metabh = ip->i_bh;
+	struct gfs2_buffer_head *metabh = ip->i_bh;
 	osi_list_t *prev_list, *cur_list, *tmp;
+	struct iptr iptr = { .ipt_ip = ip, 0};
 	int h, head_size, iblk_type;
-	uint64_t *ptr, block, *undoptr;
-	int error, was_duplicate, is_valid;
+	uint64_t *undoptr;
 	int maxptrs;
+	int error;
 
 	osi_list_add(&metabh->b_altlist, &mlp[0]);
 
@@ -1260,93 +1316,37 @@ static int build_and_check_metalist(struct gfs2_inode *ip, osi_list_t *mlp,
 		prev_list = &mlp[h - 1];
 		cur_list = &mlp[h];
 
-		for (tmp = prev_list->next; tmp != prev_list; tmp = tmp->next){
-			bh = osi_list_entry(tmp, struct gfs2_buffer_head,
-					    b_altlist);
-			if (gfs2_check_meta(bh, iblk_type)) {
+		for (tmp = prev_list->next; tmp != prev_list; tmp = tmp->next) {
+			iptr.ipt_off = head_size;
+			iptr.ipt_bh = osi_list_entry(tmp, struct gfs2_buffer_head, b_altlist);
+
+			if (gfs2_check_meta(iptr_buf(iptr), iblk_type)) {
 				if (pass->invalid_meta_is_fatal)
 					return meta_error;
 
 				continue;
 			}
-
 			if (pass->readahead)
-				file_ra(ip, bh, head_size, maxptrs, h);
+				file_ra(ip, iptr.ipt_bh, head_size, maxptrs, h);
+
 			/* Now check the metadata itself */
-			for (ptr = (uint64_t *)(bh->b_data + head_size);
-			     (char *)ptr < (bh->b_data + ip->i_sbd->bsize);
-			     ptr++) {
+			for (; iptr.ipt_off < ip->i_sbd->bsize; iptr.ipt_off += sizeof(uint64_t)) {
+				struct gfs2_buffer_head *nbh = NULL;
+
 				if (skip_this_pass || fsck_abort) {
 					free_metalist(ip, mlp);
 					return meta_is_good;
 				}
-				nbh = NULL;
-
-				if (!*ptr)
+				if (!iptr_block(iptr))
 					continue;
 
-				block = be64_to_cpu(*ptr);
-				was_duplicate = 0;
-				error = pass->check_metalist(ip, block, &nbh,
-							     h, &is_valid,
-							     &was_duplicate,
-							     pass->private);
-				/* check_metalist should hold any buffers
-				   it gets with "bread". */
-				if (error == meta_error) {
-					stack;
-					log_info(_("\nSerious metadata "
-						   "error on block %llu "
-						   "(0x%llx).\n"),
-						 (unsigned long long)block,
-						 (unsigned long long)block);
+				error = do_check_metalist(iptr, h, &nbh, pass);
+				if (error == meta_error || error == meta_skip_further)
 					goto error_undo;
-				}
-				if (error == meta_skip_further) {
-					log_info(_("\nUnrecoverable metadata "
-						   "error on block %llu "
-						   "(0x%llx). Further metadata"
-						   " will be skipped.\n"),
-						 (unsigned long long)block,
-						 (unsigned long long)block);
-					goto error_undo;
-				}
-				if (!is_valid) {
-					log_debug( _("Skipping rejected block "
-						     "%llu (0x%llx)\n"),
-						   (unsigned long long)block,
-						   (unsigned long long)block);
-					if (pass->invalid_meta_is_fatal) {
-						error = meta_error;
-						goto error_undo;
-					}
+				if (error == meta_skip_one)
 					continue;
-				}
-				/* Note that there's a special case in which
-				   we need to process the metadata block, even
-				   if it was a duplicate. That's for cases
-				   where we deleted the last reference as
-				   metadata. */
-				if (was_duplicate) {
-					log_debug( _("Skipping duplicate %llu "
-						     "(0x%llx)\n"),
-						   (unsigned long long)block,
-						   (unsigned long long)block);
-					continue;
-				}
-				if (!valid_block_ip(ip, block)) {
-					log_debug( _("Skipping invalid block "
-						     "%lld (0x%llx)\n"),
-						   (unsigned long long)block,
-						   (unsigned long long)block);
-					if (pass->invalid_meta_is_fatal) {
-						error = meta_error;
-						goto error_undo;
-					}
-					continue;
-				}
 				if (!nbh)
-					nbh = bread(ip->i_sbd, block);
+					nbh = bread(ip->i_sbd, iptr_block(iptr));
 				osi_list_add_prev(&nbh->b_altlist, cur_list);
 			} /* for all data on the indirect block */
 		} /* for blocks at that height */
@@ -1357,16 +1357,16 @@ error_undo: /* undo what we've done so far for this block */
 	if (pass->undo_check_meta == NULL)
 		return error;
 
-	log_info(_("Undoing the work we did before the error on block %llu "
-		   "(0x%llx).\n"), (unsigned long long)bh->b_blocknr,
-		 (unsigned long long)bh->b_blocknr);
-	for (undoptr = (uint64_t *)(bh->b_data + head_size); undoptr < ptr &&
-		     (char *)undoptr < (bh->b_data + ip->i_sbd->bsize);
+	log_info(_("Undoing the work we did before the error on block %"PRIu64" (0x%"PRIx64").\n"),
+	         iptr.ipt_bh->b_blocknr, iptr.ipt_bh->b_blocknr);
+	for (undoptr = (uint64_t *)(iptr_buf(iptr) + head_size);
+	     undoptr < iptr_ptr(iptr) && undoptr < iptr_endptr(iptr);
 	     undoptr++) {
-		if (!*undoptr)
+		uint64_t block = be64_to_cpu(*undoptr);
+
+		if (block == 0)
 			continue;
 
-		block = be64_to_cpu(*undoptr);
 		pass->undo_check_meta(ip, block, h, pass->private);
 	}
 	return error;
@@ -1385,9 +1385,9 @@ error_undo: /* undo what we've done so far for this block */
  *          1 if errors were found and corrected
  *          2 (ENOENT) is there were too many bad pointers
  */
-static int check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
+static int metawalk_check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
 		      struct gfs2_buffer_head *bh, int head_size,
-		      uint64_t *blks_checked, uint64_t *error_blk)
+		      uint64_t *blks_checked, struct error_block *error_blk)
 {
 	int error = 0, rc = 0;
 	uint64_t block, *ptr;
@@ -1396,8 +1396,7 @@ static int check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
 	uint64_t metablock = bh->b_blocknr;
 
 	/* If there isn't much pointer corruption check the pointers */
-	log_debug(_("\nProcessing data blocks for inode 0x%llx, metadata "
-		    "block 0x%llx.\n"),
+	log_debug("Processing data blocks for inode 0x%llx, metadata block 0x%llx.\n",
 		  (unsigned long long)ip->i_di.di_num.no_addr,
 		  (unsigned long long)bh->b_blocknr);
 	for (ptr = ptr_start ; (char *)ptr < ptr_end && !fsck_abort; ptr++) {
@@ -1418,21 +1417,36 @@ static int check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
 			log_info("\n");
 			if (rc < 0) {
 				/* A fatal error trumps a non-fatal one. */
-				if ((*error_blk == 0) || (rc < error)) {
-					log_debug(_("Fatal error on block 0x"
-						    "%llx preempts non-fatal "
-						    "error on block 0x%llx\n"),
+				if ((error_blk->errblk == 0) ||
+				    (rc < error)) {
+					log_debug(_("Fatal error on metadata "
+						    "block 0x%llx, offset "
+						    "0x%x, referencing block "
+						    "0x%llx preempts non-fatal"
+						    " error on block 0x%llx\n"),
+						  (unsigned long long)metablock,
+						  (int)(ptr - ptr_start),
 						  (unsigned long long)block,
-						  (unsigned long long)*error_blk);
-					*error_blk = block;
+						  (unsigned long long)error_blk->errblk);
+					error_blk->metablk = metablock;
+					error_blk->metaoff = ptr - ptr_start;
+					error_blk->errblk = block;
 				}
 				log_info(_("Unrecoverable "));
 			} else { /* nonfatal error */
-				if ((*error_blk) == 0)
-					*error_blk = block;
+				if (error_blk->errblk == 0) {
+					error_blk->metablk = metablock;
+					error_blk->metaoff = ptr - ptr_start;
+					error_blk->errblk = block;
+				}
 			}
-			log_info(_("data block error %d on block %llu "
-				   "(0x%llx).\n"), rc,
+			log_info(_("data block error %d on metadata block "
+				   "%lld (0x%llx), offset %d (0x%x), "
+				   "referencing data block %lld (0x%llx).\n"),
+				 rc, (unsigned long long)metablock,
+				 (unsigned long long)metablock,
+				 (int)(ptr - ptr_start),
+				 (int)(ptr - ptr_start),
 				 (unsigned long long)block,
 				 (unsigned long long)block);
 			error = rc;
@@ -1445,8 +1459,9 @@ static int check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
 }
 
 static int undo_check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
+			   uint64_t metablock,
 			   uint64_t *ptr_start, char *ptr_end,
-			   uint64_t error_blk, int error)
+			   struct error_block *error_blk, int error)
 {
 	int rc = 0;
 	uint64_t block, *ptr;
@@ -1460,19 +1475,27 @@ static int undo_check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
 		if (skip_this_pass || fsck_abort)
 			return 1;
 		block =  be64_to_cpu(*ptr);
-		if (block == error_blk) {
+		if (metablock == error_blk->metablk &&
+		    (ptr - ptr_start == error_blk->metaoff) &&
+		    block == error_blk->errblk) {
 			if (error < 0) { /* A fatal error that stopped it? */
 				log_debug(_("Stopping the undo process: "
 					    "fatal error block 0x%llx was "
-					    "found.\n"),
-					  (unsigned long long)error_blk);
+					    "found at metadata block 0x%llx,"
+					    "offset 0x%x.\n"),
+					  (unsigned long long)error_blk->errblk,
+					  (unsigned long long)error_blk->metablk,
+					  error_blk->metaoff);
 				return 1;
 			}
 			found_error_blk = 1;
 			log_debug(_("The non-fatal error block 0x%llx was "
-				    "found, but undo processing will continue "
+				    "found at metadata block 0x%llx, offset "
+				    "0x%d, but undo processing will continue "
 				    "until the end of this metadata block.\n"),
-				  (unsigned long long)error_blk);
+				  (unsigned long long)error_blk->errblk,
+				  (unsigned long long)error_blk->metablk,
+				  error_blk->metaoff);
 		}
 		rc = pass->undo_check_data(ip, block, pass->private);
 		if (rc < 0)
@@ -1484,7 +1507,7 @@ static int undo_check_data(struct gfs2_inode *ip, struct metawalk_fxns *pass,
 static int hdr_size(struct gfs2_buffer_head *bh, int height)
 {
 	if (height > 1) {
-		if (gfs2_check_meta(bh, GFS2_METATYPE_IN))
+		if (gfs2_check_meta(bh->b_data, GFS2_METATYPE_IN))
 			return 0;
 		if (bh->sdp->gfs1)
 			return sizeof(struct gfs_indirect);
@@ -1492,7 +1515,7 @@ static int hdr_size(struct gfs2_buffer_head *bh, int height)
 			return sizeof(struct gfs2_meta_header);
 	}
 	/* if this isn't really a dinode, skip it */
-	if (gfs2_check_meta(bh, GFS2_METATYPE_DI))
+	if (gfs2_check_meta(bh->b_data, GFS2_METATYPE_DI))
 		return 0;
 
 	return sizeof(struct gfs2_dinode);
@@ -1514,7 +1537,7 @@ int check_metatree(struct gfs2_inode *ip, struct metawalk_fxns *pass)
 	uint64_t blks_checked = 0;
 	int error, rc;
 	int metadata_clean = 0;
-	uint64_t error_blk = 0;
+	struct error_block error_blk = {0, 0, 0};
 	int hit_error_blk = 0;
 
 	if (!height && !is_dir(&ip->i_di, ip->i_sbd->gfs1))
@@ -1560,7 +1583,7 @@ int check_metatree(struct gfs2_inode *ip, struct metawalk_fxns *pass)
 			continue;
 
 		if (pass->check_data)
-			error = check_data(ip, pass, bh, head_size,
+			error = metawalk_check_data(ip, pass, bh, head_size,
 					   &blks_checked, &error_blk);
 		if (pass->big_file_msg && ip->i_di.di_blocks > COMFORTABLE_BLKS)
 			pass->big_file_msg(ip, blks_checked);
@@ -1577,10 +1600,15 @@ undo_metalist:
 	if (!error)
 		goto out;
 	log_err( _("Error: inode %llu (0x%llx) had unrecoverable errors at "
+		   "metadata block %lld (0x%llx), offset %d (0x%x), block "
 		   "%lld (0x%llx).\n"),
 		 (unsigned long long)ip->i_di.di_num.no_addr,
 		 (unsigned long long)ip->i_di.di_num.no_addr,
-		 (unsigned long long)error_blk, (unsigned long long)error_blk);
+		 (unsigned long long)error_blk.metablk,
+		 (unsigned long long)error_blk.metablk,
+		 error_blk.metaoff, error_blk.metaoff,
+		 (unsigned long long)error_blk.errblk,
+		 (unsigned long long)error_blk.errblk);
 	if (!query( _("Remove the invalid inode? (y/n) "))) {
 		free_metalist(ip, &metalist[0]);
 		log_err(_("Invalid inode not deleted.\n"));
@@ -1606,12 +1634,20 @@ undo_metalist:
 				head_size = hdr_size(bh, height);
 				if (head_size) {
 					rc = undo_check_data(ip, pass,
+							     bh->b_blocknr,
 							     (uint64_t *)
 					      (bh->b_data + head_size),
 					      (bh->b_data + ip->i_sbd->bsize),
-							     error_blk, error);
+							     &error_blk,
+							     error);
 					if (rc > 0) {
 						hit_error_blk = 1;
+						log_err("Reached the error "
+							"block undoing work "
+							"for inode %lld "
+							"(0x%llx).\n",
+							(unsigned long long)ip->i_di.di_num.no_addr,
+							(unsigned long long)ip->i_di.di_num.no_addr);
 						rc = 0;
 					}
 				}
@@ -1628,7 +1664,7 @@ undo_metalist:
 	   to undo. */
 	delete_all_dups(ip);
 	/* Set the dinode as "bad" so it gets deleted */
-	fsck_bitmap_set(ip, ip->i_di.di_num.no_addr, _("corrupt"),
+	fsck_bitmap_set(ip, ip->i_di.di_num.no_addr, "corrupt",
 			GFS2_BLKST_FREE);
 	log_err(_("The corrupt inode was invalidated.\n"));
 out:
