@@ -39,12 +39,11 @@ struct block_count {
 };
 
 static int p1check_leaf(struct gfs2_inode *ip, uint64_t block, void *private);
-static int check_metalist(struct gfs2_inode *ip, uint64_t block,
-			  struct gfs2_buffer_head **bh, int h, int *is_valid,
-			  int *was_duplicate, void *private);
+static int pass1_check_metalist(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                                int *is_valid, int *was_duplicate, void *private);
 static int undo_check_metalist(struct gfs2_inode *ip, uint64_t block,
 			       int h, void *private);
-static int check_data(struct gfs2_inode *ip, uint64_t metablock,
+static int pass1_check_data(struct gfs2_inode *ip, uint64_t metablock,
 		      uint64_t block, void *private,
 		      struct gfs2_buffer_head *bh, uint64_t *ptr);
 static int undo_check_data(struct gfs2_inode *ip, uint64_t block,
@@ -69,10 +68,8 @@ static int check_extended_leaf_eattr(struct gfs2_inode *ip, int i,
 				     void *private);
 static int finish_eattr_indir(struct gfs2_inode *ip, int leaf_pointers,
 			      int leaf_pointer_errors, void *private);
-static int invalidate_metadata(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, int h,
-			       int *is_valid, int *was_duplicate,
-			       void *private);
+static int invalidate_metadata(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                               int *is_valid, int *was_duplicate, void *private);
 static int invalidate_leaf(struct gfs2_inode *ip, uint64_t block,
 			   void *private);
 static int invalidate_data(struct gfs2_inode *ip, uint64_t metablock,
@@ -191,8 +188,8 @@ out:
 struct metawalk_fxns pass1_fxns = {
 	.private = NULL,
 	.check_leaf = p1check_leaf,
-	.check_metalist = check_metalist,
-	.check_data = check_data,
+	.check_metalist = pass1_check_metalist,
+	.check_data = pass1_check_data,
 	.check_eattr_indir = check_eattr_indir,
 	.check_eattr_leaf = check_eattr_leaf,
 	.check_dentry = NULL,
@@ -223,12 +220,12 @@ struct metawalk_fxns invalidate_fxns = {
  * marked "in use" by the bitmap.  You don't want root's indirect blocks
  * deleted, do you? Or worse, reused for lost+found.
  */
-static int resuscitate_metalist(struct gfs2_inode *ip, uint64_t block,
-				struct gfs2_buffer_head **bh, int h,
-				int *is_valid, int *was_duplicate,
-				void *private)
+static int resuscitate_metalist(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                                int *is_valid, int *was_duplicate, void *private)
 {
 	struct block_count *bc = (struct block_count *)private;
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
 
 	*is_valid = 1;
 	*was_duplicate = 0;
@@ -344,15 +341,16 @@ static int p1check_leaf(struct gfs2_inode *ip, uint64_t block, void *private)
 	return 0;
 }
 
-static int check_metalist(struct gfs2_inode *ip, uint64_t block,
-			  struct gfs2_buffer_head **bh, int h, int *is_valid,
-			  int *was_duplicate, void *private)
+static int pass1_check_metalist(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                                int *is_valid, int *was_duplicate, void *private)
 {
-	int q;
-	int iblk_type;
-	struct gfs2_buffer_head *nbh;
 	struct block_count *bc = (struct block_count *)private;
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
+	struct gfs2_buffer_head *nbh;
 	const char *blktypedesc;
+	int iblk_type;
+	int q;
 
 	*bh = NULL;
 
@@ -393,7 +391,7 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 	}
 	nbh = bread(ip->i_sbd, block);
 
-	*is_valid = (gfs2_check_meta(nbh, iblk_type) == 0);
+	*is_valid = (gfs2_check_meta(nbh->b_data, iblk_type) == 0);
 
 	if (!(*is_valid)) {
 		log_err( _("Inode %lld (0x%llx) has a bad indirect block "
@@ -403,8 +401,15 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 			 (unsigned long long)ip->i_di.di_num.no_addr,
 			 (unsigned long long)block,
 			 (unsigned long long)block, blktypedesc);
-		brelse(nbh);
-		return meta_skip_further;
+		if (query(_("Zero the indirect block pointer? (y/n) "))){
+			*iptr_ptr(iptr) = 0;
+			bmodified(iptr.ipt_bh);
+			*is_valid = 1;
+			return meta_skip_one;
+		} else {
+			brelse(nbh);
+			return meta_skip_further;
+		}
 	}
 
 	bc->indir_count++;
@@ -511,7 +516,7 @@ static int blockmap_set_as_data(struct gfs2_inode *ip, uint64_t block)
 	struct gfs2_buffer_head *bh;
 	struct gfs2_dinode *di;
 
-	error = fsck_blkmap_set_noino(ip, block, _("data"), GFS2_BLKST_USED);
+	error = fsck_blkmap_set_noino(ip, block, "data", GFS2_BLKST_USED);
 	if (!error)
 		return 0;
 
@@ -519,7 +524,7 @@ static int blockmap_set_as_data(struct gfs2_inode *ip, uint64_t block)
 	/* The bitmap says it's a dinode, but a block reference begs to differ.
 	   So which is it? */
 	bh = bread(ip->i_sbd, block);
-	if (gfs2_check_meta(bh, GFS2_METATYPE_DI) != 0)
+	if (gfs2_check_meta(bh->b_data, GFS2_METATYPE_DI) != 0)
 		goto out;
 
 	/* The meta header agrees it's a dinode. But it might be data in
@@ -537,12 +542,12 @@ static int blockmap_set_as_data(struct gfs2_inode *ip, uint64_t block)
 	error = -1;
 out:
 	if (!error)
-		fsck_blockmap_set(ip, block, _("data"),  GFS2_BLKST_USED);
+		fsck_blockmap_set(ip, block, "data",  GFS2_BLKST_USED);
 	brelse(bh);
 	return error;
 }
 
-static int check_data(struct gfs2_inode *ip, uint64_t metablock,
+static int pass1_check_data(struct gfs2_inode *ip, uint64_t metablock,
 		      uint64_t block, void *private,
 		      struct gfs2_buffer_head *bbh, uint64_t *ptr)
 {
@@ -642,13 +647,13 @@ static int check_data(struct gfs2_inode *ip, uint64_t metablock,
 		log_info(_("Block %lld (0x%llx) is a GFS1 rindex block\n"),
 			 (unsigned long long)block, (unsigned long long)block);
 		gfs2_special_set(&gfs1_rindex_blks, block);
-		fsck_blockmap_set(ip, block, _("rgrp"), GFS2_BLKST_DINODE);
+		fsck_blockmap_set(ip, block, "rgrp", GFS2_BLKST_DINODE);
 		/*gfs2_meta_rgrp);*/
 	} else if (ip->i_sbd->gfs1 && ip->i_di.di_flags & GFS2_DIF_JDATA) {
 		log_info(_("Block %lld (0x%llx) is a GFS1 journaled data "
 			   "block\n"),
 			 (unsigned long long)block, (unsigned long long)block);
-		fsck_blockmap_set(ip, block, _("jdata"), GFS2_BLKST_DINODE);
+		fsck_blockmap_set(ip, block, "jdata", GFS2_BLKST_DINODE);
 	} else
 		return blockmap_set_as_data(ip, block);
 	return 0;
@@ -747,7 +752,7 @@ static int check_eattr_indir(struct gfs2_inode *ip, uint64_t indirect,
 	   handling sort it out.  If it isn't, clear it but don't
 	   count it as a duplicate. */
 	*bh = bread(sdp, indirect);
-	if (gfs2_check_meta(*bh, GFS2_METATYPE_IN)) {
+	if (gfs2_check_meta((*bh)->b_data, GFS2_METATYPE_IN)) {
 		bc->ea_count++;
 		if (q != GFS2_BLKST_FREE) { /* Duplicate? */
 			add_duplicate_ref(ip, indirect, ref_as_ea, 0,
@@ -829,7 +834,7 @@ static int check_ealeaf_block(struct gfs2_inode *ip, uint64_t block, int btype,
 	   really is an EA.  If it is, let duplicate handling sort it out.
 	   If it isn't, clear it but don't count it as a duplicate. */
 	leaf_bh = bread(sdp, block);
-	if (gfs2_check_meta(leaf_bh, btype)) {
+	if (gfs2_check_meta(leaf_bh->b_data, btype)) {
 		bc->ea_count++;
 		if (q != GFS2_BLKST_FREE) { /* Duplicate? */
 			add_duplicate_ref(ip, block, ref_as_ea, 0,
@@ -1111,11 +1116,12 @@ static int mark_block_invalid(struct gfs2_inode *ip, uint64_t block,
 	return meta_is_good;
 }
 
-static int invalidate_metadata(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, int h,
-			       int *is_valid, int *was_duplicate,
-			       void *private)
+static int invalidate_metadata(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                               int *is_valid, int *was_duplicate, void *private)
 {
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
+
 	*is_valid = 1;
 	*was_duplicate = 0;
 	return mark_block_invalid(ip, block, ref_as_meta, _("metadata"),
@@ -1214,11 +1220,12 @@ static int rangecheck_block(struct gfs2_inode *ip, uint64_t block,
 	return meta_is_good;
 }
 
-static int rangecheck_metadata(struct gfs2_inode *ip, uint64_t block,
-			       struct gfs2_buffer_head **bh, int h,
-			       int *is_valid, int *was_duplicate,
-			       void *private)
+static int rangecheck_metadata(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                               int *is_valid, int *was_duplicate, void *private)
 {
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
+
 	*is_valid = 1;
 	*was_duplicate = 0;
 	return rangecheck_block(ip, block, bh, btype_meta, private);
@@ -1286,25 +1293,25 @@ static int set_ip_blockmap(struct gfs2_inode *ip)
 
 	switch (mode) {
 	case S_IFDIR:
-		ty = _("directory");
+		ty = "directory";
 		break;
 	case S_IFREG:
-		ty = _("file");
+		ty = "file";
 		break;
 	case S_IFLNK:
-		ty = _("symlink");
+		ty = "symlink";
 		break;
 	case S_IFBLK:
-		ty = _("block device");
+		ty = "block device";
 		break;
 	case S_IFCHR:
-		ty = _("character device");
+		ty = "character device";
 		break;
 	case S_IFIFO:
-		ty = _("fifo");
+		ty = "fifo";
 		break;
 	case S_IFSOCK:
-		ty = _("socket");
+		ty = "socket";
 		break;
 	default:
 		return -EINVAL;
@@ -1317,12 +1324,13 @@ static int set_ip_blockmap(struct gfs2_inode *ip)
 	return 0;
 }
 
-static int alloc_metalist(struct gfs2_inode *ip, uint64_t block,
-			  struct gfs2_buffer_head **bh, int h, int *is_valid,
-			  int *was_duplicate, void *private)
+static int alloc_metalist(struct iptr iptr, struct gfs2_buffer_head **bh, int h,
+                          int *is_valid, int *was_duplicate, void *private)
 {
-	int q;
 	const char *desc = (const char *)private;
+	struct gfs2_inode *ip = iptr.ipt_ip;
+	uint64_t block = iptr_block(iptr);
+	int q;
 
 	/* No need to range_check here--if it was added, it's in range. */
 	/* We can't check the bitmap here because this function is called
@@ -1372,7 +1380,7 @@ static int alloc_leaf(struct gfs2_inode *ip, uint64_t block, void *private)
 	   after the bitmap has been set but before the blockmap has. */
 	q = bitmap_type(ip->i_sbd, block);
 	if (q == GFS2_BLKST_FREE)
-		fsck_blockmap_set(ip, block, _("newly allocated leaf"),
+		fsck_blockmap_set(ip, block, "newly allocated leaf",
 				  ip->i_sbd->gfs1 ? GFS2_BLKST_DINODE :
 				  GFS2_BLKST_USED);
 	return 0;
@@ -1666,7 +1674,7 @@ static int check_system_inode(struct gfs2_sbd *sdp,
 			 " (0x%llx)\n"), filename,
 			 (unsigned long long)iblock,
 			 (unsigned long long)iblock);
-		if (gfs2_check_meta((*sysinode)->i_bh, GFS2_METATYPE_DI)) {
+		if (gfs2_check_meta((*sysinode)->i_bh->b_data, GFS2_METATYPE_DI)) {
 			log_err( _("Found invalid system dinode at block #"
 				   "%llu (0x%llx)\n"),
 				 (unsigned long long)iblock,
@@ -1935,7 +1943,7 @@ static int pass1_process_bitmap(struct gfs2_sbd *sdp, struct rgrp_tree *rgd, uin
 		bh = bread(sdp, block);
 
 		is_inode = 0;
-		if (gfs2_check_meta(bh, GFS2_METATYPE_DI) == 0)
+		if (gfs2_check_meta(bh->b_data, GFS2_METATYPE_DI) == 0)
 			is_inode = 1;
 
 		check_magic = ((struct gfs2_meta_header *)
@@ -2193,12 +2201,11 @@ int pass1(struct gfs2_sbd *sdp)
 			goto out;
 		}
 		next = osi_next(n);
-		log_debug( _("Checking metadata in Resource Group #%llu\n"),
+		log_debug("Checking metadata in resource group #%llu\n",
 				 (unsigned long long)rg_count);
 		rgd = (struct rgrp_tree *)n;
 		for (i = 0; i < rgd->ri.ri_length; i++) {
-			log_debug( _("rgrp block %lld (0x%llx) "
-				     "is now marked as 'rgrp data'\n"),
+			log_debug("rgrp block %lld (0x%llx) is now marked as 'rgrp data'\n",
 				   rgd->ri.ri_addr + i, rgd->ri.ri_addr + i);
 			if (gfs2_blockmap_set(bl, rgd->ri.ri_addr + i,
 					      GFS2_BLKST_USED)) {
