@@ -25,13 +25,12 @@ static __inline__ int fs_is_jdata(struct gfs2_inode *ip)
 }
 
 static __inline__ uint64_t *
-gfs1_metapointer(struct gfs2_buffer_head *bh, unsigned int height,
-		 struct metapath *mp)
+gfs1_metapointer(char *buf, unsigned int height, struct metapath *mp)
 {
 	unsigned int head_size = (height > 0) ?
 		sizeof(struct gfs_indirect) : sizeof(struct gfs_dinode);
 
-	return ((uint64_t *)(bh->b_data + head_size)) + mp->mp_list[height];
+	return ((uint64_t *)(buf + head_size)) + mp->mp_list[height];
 }
 
 int is_gfs_dir(struct gfs2_dinode *dinode)
@@ -45,7 +44,7 @@ void gfs1_lookup_block(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 		  unsigned int height, struct metapath *mp,
 		  int create, int *new, uint64_t *block)
 {
-	uint64_t *ptr = gfs1_metapointer(bh, height, mp);
+	uint64_t *ptr = gfs1_metapointer(bh->b_data, height, mp);
 
 	if (*ptr) {
 		*block = be64_to_cpu(*ptr);
@@ -248,14 +247,11 @@ int gfs1_writei(struct gfs2_inode *ip, char *buf, uint64_t offset,
 	return copied;
 }
 
-/* ------------------------------------------------------------------------ */
-/* gfs_dinode_in */
-/* ------------------------------------------------------------------------ */
-static void gfs_dinode_in(struct gfs_dinode *di, struct gfs2_buffer_head *bh)
+static void gfs_dinode_in(struct gfs_dinode *di, char *buf)
 {
-	struct gfs_dinode *str = (struct gfs_dinode *)bh->b_data;
+	struct gfs_dinode *str = (struct gfs_dinode *)buf;
 
-	gfs2_meta_header_in(&di->di_header, bh->b_data);
+	gfs2_meta_header_in(&di->di_header, buf);
 	gfs2_inum_in(&di->di_num, (char *)&str->di_num);
 
 	di->di_mode = be32_to_cpu(str->di_mode);
@@ -280,9 +276,7 @@ static void gfs_dinode_in(struct gfs_dinode *di, struct gfs2_buffer_head *bh)
 	di->di_eattr = be64_to_cpu(str->di_eattr);
 }
 
-static struct gfs2_inode *__gfs_inode_get(struct gfs2_sbd *sdp,
-					  struct gfs2_buffer_head *bh,
-					  uint64_t di_addr)
+static struct gfs2_inode *__gfs_inode_get(struct gfs2_sbd *sdp, char *buf, uint64_t di_addr)
 {
 	struct gfs_dinode gfs1_dinode;
 	struct gfs2_inode *ip;
@@ -291,13 +285,7 @@ static struct gfs2_inode *__gfs_inode_get(struct gfs2_sbd *sdp,
 	if (ip == NULL) {
 		return NULL;
 	}
-
-	ip->bh_owned = 0;
-	if (!bh) {
-		bh = bread(sdp, di_addr);
-		ip->bh_owned = 1;
-	}
-	gfs_dinode_in(&gfs1_dinode, bh);
+	gfs_dinode_in(&gfs1_dinode, buf);
 	memcpy(&ip->i_di.di_header, &gfs1_dinode.di_header,
 	       sizeof(struct gfs2_meta_header));
 	memcpy(&ip->i_di.di_num, &gfs1_dinode.di_num,
@@ -322,20 +310,31 @@ static struct gfs2_inode *__gfs_inode_get(struct gfs2_sbd *sdp,
 	ip->i_di.di_depth = gfs1_dinode.di_depth;
 	ip->i_di.di_entries = gfs1_dinode.di_entries;
 	ip->i_di.di_eattr = gfs1_dinode.di_eattr;
-	ip->i_bh = bh;
 	ip->i_sbd = sdp;
 	return ip;
 }
 
-struct gfs2_inode *lgfs2_gfs_inode_get(struct gfs2_sbd *sdp,
-				 struct gfs2_buffer_head *bh)
+struct gfs2_inode *lgfs2_gfs_inode_get(struct gfs2_sbd *sdp, char *buf)
 {
-	return __gfs_inode_get(sdp, bh, 0);
+	return __gfs_inode_get(sdp, buf, 0);
 }
 
 struct gfs2_inode *lgfs2_gfs_inode_read(struct gfs2_sbd *sdp, uint64_t di_addr)
 {
-	return __gfs_inode_get(sdp, NULL, di_addr);
+	struct gfs2_buffer_head *bh;
+	struct gfs2_inode *ip;
+
+	bh = bget(sdp, di_addr);
+	if (bh == NULL)
+		return NULL;
+	if (pread(sdp->device_fd, bh->b_data, sdp->bsize, di_addr * sdp->bsize) != sdp->bsize) {
+		brelse(bh);
+		return NULL;
+	}
+	ip = __gfs_inode_get(sdp, bh->b_data, di_addr);
+	ip->i_bh = bh;
+	ip->bh_owned = 1;
+	return ip;
 }
 
 /* ------------------------------------------------------------------------ */
@@ -351,14 +350,11 @@ void gfs_jindex_in(struct gfs_jindex *jindex, char *jbuf)
 	memcpy(jindex->ji_reserved, str->ji_reserved, 64);
 }
 
-/* ------------------------------------------------------------------------ */
-/* gfs_rgrp_in - Read in a resource group header                            */
-/* ------------------------------------------------------------------------ */
-void gfs_rgrp_in(struct gfs_rgrp *rgrp, struct gfs2_buffer_head *rbh)
+void gfs_rgrp_in(struct gfs_rgrp *rgrp, const char *buf)
 {
-	struct gfs_rgrp *str = (struct gfs_rgrp *)rbh->b_data;
+	struct gfs_rgrp *str = (struct gfs_rgrp *)buf;
 
-	gfs2_meta_header_in(&rgrp->rg_header, rbh->b_data);
+	gfs2_meta_header_in(&rgrp->rg_header, buf);
 	rgrp->rg_flags = be32_to_cpu(str->rg_flags);
 	rgrp->rg_free = be32_to_cpu(str->rg_free);
 	rgrp->rg_useddi = be32_to_cpu(str->rg_useddi);
@@ -370,14 +366,11 @@ void gfs_rgrp_in(struct gfs_rgrp *rgrp, struct gfs2_buffer_head *rbh)
 	memcpy(rgrp->rg_reserved, str->rg_reserved, 64);
 }
 
-/* ------------------------------------------------------------------------ */
-/* gfs_rgrp_out */
-/* ------------------------------------------------------------------------ */
-void gfs_rgrp_out(struct gfs_rgrp *rgrp, struct gfs2_buffer_head *rbh)
+void gfs_rgrp_out(const struct gfs_rgrp *rgrp, char *buf)
 {
-	struct gfs_rgrp *str = (struct gfs_rgrp *)rbh->b_data;
+	struct gfs_rgrp *str = (struct gfs_rgrp *)buf;
 
-	gfs2_meta_header_out(&rgrp->rg_header, rbh->b_data);
+	gfs2_meta_header_out(&rgrp->rg_header, buf);
 	str->rg_flags = cpu_to_be32(rgrp->rg_flags);
 	str->rg_free = cpu_to_be32(rgrp->rg_free);
 	str->rg_useddi = cpu_to_be32(rgrp->rg_useddi);
@@ -387,5 +380,4 @@ void gfs_rgrp_out(struct gfs_rgrp *rgrp, struct gfs2_buffer_head *rbh)
 	str->rg_freemeta = cpu_to_be32(rgrp->rg_freemeta);
 
 	memcpy(str->rg_reserved, rgrp->rg_reserved, 64);
-	bmodified(rbh);
 }
