@@ -156,44 +156,68 @@ int lgfs2_write_journal_data(struct gfs2_inode *ip)
 		.lh_flags = GFS2_LOG_HEAD_UNMOUNT,
 #endif
 	};
-	struct gfs2_buffer_head *bh;
 	struct gfs2_sbd *sdp = ip->i_sbd;
 	unsigned blocks = (ip->i_di.di_size + sdp->bsize - 1) / sdp->bsize;
 	uint64_t jext0 = ip->i_di.di_num.no_addr + ip->i_di.di_blocks - blocks;
 	uint64_t seq = ((blocks) * (random() / (RAND_MAX + 1.0)));
+	uint64_t jblk = jext0;
+	char *buf;
 
-	bh = bget(sdp, jext0);
-	if (bh == NULL)
+	buf = calloc(1, sdp->bsize);
+	if (buf == NULL)
 		return -1;
 
 	crc32c_optimization_init();
 	do {
-		struct gfs2_log_header *buflh = (struct gfs2_log_header *)bh->b_data;
+		struct gfs2_log_header *buflh = (struct gfs2_log_header *)buf;
 
 		lh.lh_sequence = seq;
-		lh.lh_blkno = bh->b_blocknr - jext0;
-		gfs2_log_header_out(&lh, bh->b_data);
+		lh.lh_blkno = jblk - jext0;
+		gfs2_log_header_out(&lh, buf);
 
-		buflh->lh_hash = cpu_to_be32(lgfs2_log_header_hash(bh->b_data));
+		buflh->lh_hash = cpu_to_be32(lgfs2_log_header_hash(buf));
 #ifdef GFS2_HAS_LH_V2
-		buflh->lh_addr = cpu_to_be64(bh->b_blocknr);
-		buflh->lh_crc = cpu_to_be32(lgfs2_log_header_crc(bh->b_data, sdp->bsize));
+		buflh->lh_addr = cpu_to_be64(jblk);
+		buflh->lh_crc = cpu_to_be32(lgfs2_log_header_crc(buf, sdp->bsize));
 #endif
-
-		if (bwrite(bh)) {
-			free(bh);
+		if (pwrite(sdp->device_fd, buf, sdp->bsize, jblk * sdp->bsize) != sdp->bsize) {
+			free(buf);
 			return -1;
 		}
 
 		if (++seq == blocks)
 			seq = 0;
 
-	} while (++bh->b_blocknr < jext0 + blocks);
+	} while (++jblk < jext0 + blocks);
 
-	free(bh);
+	free(buf);
 	return 0;
 }
 
+static struct gfs2_buffer_head *get_file_buf(struct gfs2_inode *ip, uint64_t lbn, int prealloc)
+{
+	struct gfs2_sbd *sdp = ip->i_sbd;
+	uint64_t dbn;
+	int new = TRUE;
+
+	if (ip->i_di.di_height == 0)
+		unstuff_dinode(ip);
+
+	block_map(ip, lbn, &new, &dbn, NULL, prealloc);
+	if (!dbn)
+		return NULL;
+
+	if (!prealloc && new &&
+	    ip->i_di.di_size < (lbn + 1) << sdp->sd_sb.sb_bsize_shift) {
+		bmodified(ip->i_bh);
+		ip->i_di.di_size = (lbn + 1) << sdp->sd_sb.sb_bsize_shift;
+	}
+	if (dbn == ip->i_di.di_num.no_addr)
+		return ip->i_bh;
+	else
+		return bread(sdp, dbn);
+}
+
 int write_journal(struct gfs2_inode *jnl, unsigned bsize, unsigned int blocks)
 {
 	struct gfs2_log_header lh;
@@ -648,10 +672,10 @@ int do_init_statfs(struct gfs2_sbd *sdp)
 	return 0;
 }
 
-int gfs2_check_meta(struct gfs2_buffer_head *bh, int type)
+int gfs2_check_meta(const char *buf, int type)
 {
-	uint32_t check_magic = ((struct gfs2_meta_header *)(bh->b_data))->mh_magic;
-	uint32_t check_type = ((struct gfs2_meta_header *)(bh->b_data))->mh_type;
+	uint32_t check_magic = ((struct gfs2_meta_header *)buf)->mh_magic;
+	uint32_t check_type = ((struct gfs2_meta_header *)buf)->mh_type;
 
 	check_magic = be32_to_cpu(check_magic);
 	check_type = be32_to_cpu(check_type);
@@ -667,13 +691,12 @@ unsigned lgfs2_bm_scan(struct rgrp_tree *rgd, unsigned idx, uint64_t *buf, uint8
 	uint32_t blk = 0;
 
 	while(blk < (bi->bi_len * GFS2_NBBY)) {
-		blk = gfs2_bitfit((uint8_t *)bi->bi_bh->b_data + bi->bi_offset,
+		blk = gfs2_bitfit((uint8_t *)bi->bi_data + bi->bi_offset,
 				  bi->bi_len, blk, state);
 		if (blk == BFITNOENT)
 			break;
 		buf[n++] = blk + (bi->bi_start * GFS2_NBBY) + rgd->ri.ri_data0;
 		blk++;
 	}
-
 	return n;
 }
