@@ -311,7 +311,8 @@ static void opt_parse_extended(char *str, struct mkfs_opts *opts)
 			print_ext_opts();
 			exit(0);
 		} else {
-			fprintf(stderr, _("Invalid option '%s'\n"), key);
+			fprintf(stderr, _("Invalid extended option (specified with -o): '%s'\n"), key);
+			print_ext_opts();
 			exit(-1);
 		}
 	}
@@ -505,7 +506,7 @@ static unsigned choose_blocksize(struct mkfs_opts *opts)
 	}
 	if (!opts->got_bsize && got_topol) {
 		if (dev->optimal_io_size <= getpagesize() &&
-		    dev->optimal_io_size >= dev->minimum_io_size)
+		    dev->optimal_io_size >= GFS2_DEFAULT_BSIZE)
 			bsize = dev->optimal_io_size;
 		else if (dev->physical_sector_size <= getpagesize() &&
 		         dev->physical_sector_size >= GFS2_DEFAULT_BSIZE)
@@ -710,9 +711,16 @@ static int place_rgrp(struct gfs2_sbd *sdp, lgfs2_rgrp_t rg, int debug)
 		prev_end = lgfs2_rgrp_index(prev)->ri_data0 +
 		           lgfs2_rgrp_index(prev)->ri_data;
 	}
-	err = zero_gap(sdp, prev_end, ri->ri_addr - prev_end);
-	if (err != 0)
-		return -1;
+	while (prev_end < ri->ri_addr) {
+		size_t gap_len = ri->ri_addr - prev_end;
+
+		if (gap_len > IOV_MAX)
+			gap_len = IOV_MAX;
+		err = zero_gap(sdp, prev_end, gap_len);
+		if (err != 0)
+			return -1;
+		prev_end += gap_len;
+	}
 	err = lgfs2_rgrp_write(sdp->device_fd, rg);
 	if (err != 0) {
 		perror(_("Failed to write resource group"));
@@ -921,11 +929,17 @@ static void sbd_init(struct gfs2_sbd *sdp, struct mkfs_opts *opts, unsigned bsiz
 	   will fit. For user-provided journal sizes, limit it to half of the fs. */
 	if (!opts->got_jsize) {
 		int default_jsize = default_journal_size(sdp->bsize, sdp->device.length / opts->journals);
+		unsigned jsize_mb;
+
 		if (default_jsize < 0) {
 			fprintf(stderr, _("gfs2 will not fit on this device.\n"));
 			exit(1);
 		}
-		opts->jsize = (default_jsize * sdp->bsize) >> 20;
+		jsize_mb = (default_jsize * sdp->bsize) >> 20;
+		if (jsize_mb < GFS2_MIN_JSIZE)
+			opts->jsize = GFS2_MIN_JSIZE;
+		else
+			opts->jsize = jsize_mb;
 	} else if ((((opts->jsize * opts->journals) << 20) / sdp->bsize) > (sdp->device.length / 2)) {
 		unsigned max_jsize = (sdp->device.length / 2 * sdp->bsize / opts->journals) >> 20;
 
@@ -1143,8 +1157,8 @@ int main(int argc, char *argv[])
 	build_root(&sbd);
 	sb.sb_root_dir = sbd.md.rooti->i_di.di_num;
 
-	strncpy(sb.sb_lockproto, opts.lockproto, GFS2_LOCKNAME_LEN);
-	strncpy(sb.sb_locktable, opts.locktable, GFS2_LOCKNAME_LEN);
+	strncpy(sb.sb_lockproto, opts.lockproto, GFS2_LOCKNAME_LEN - 1);
+	strncpy(sb.sb_locktable, opts.locktable, GFS2_LOCKNAME_LEN - 1);
 	sb.sb_lockproto[GFS2_LOCKNAME_LEN - 1] = '\0';
 	sb.sb_locktable[GFS2_LOCKNAME_LEN - 1] = '\0';
 
